# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gitlab
from gitlab import base
from gitlab import cli
from gitlab import exceptions as exc
from gitlab import types as g_types
from gitlab import utils


class GetMixin(object):
    @exc.on_http_error(exc.GitlabGetError)
    def get(self, id, lazy=False, **kwargs):
        """Retrieve a single object.

        Args:
            id (int or str): ID of the object to retrieve
            lazy (bool): If True, don't request the server, but create a
                         shallow object giving access to the managers. This is
                         useful if you want to avoid useless calls to the API.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            object: The generated RESTObject.

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request
        """
        if not isinstance(id, int):
            id = utils.clean_str_id(id)
        path = "%s/%s" % (self.path, id)
        if lazy is True:
            return self._obj_cls(self, {self._obj_cls._id_attr: id})
        server_data = self.gitlab.http_get(path, **kwargs)
        return self._obj_cls(self, server_data)


class GetWithoutIdMixin(object):
    @exc.on_http_error(exc.GitlabGetError)
    def get(self, id=None, **kwargs):
        """Retrieve a single object.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            object: The generated RESTObject

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request
        """
        server_data = self.gitlab.http_get(self.path, **kwargs)
        if server_data is None:
            return None
        return self._obj_cls(self, server_data)


class RefreshMixin(object):
    @exc.on_http_error(exc.GitlabGetError)
    def refresh(self, **kwargs):
        """Refresh a single object from server.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns None (updates the object)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request
        """
        if self._id_attr:
            path = "%s/%s" % (self.manager.path, self.id)
        else:
            path = self.manager.path
        server_data = self.manager.gitlab.http_get(path, **kwargs)
        self._update_attrs(server_data)


class ListMixin(object):
    @exc.on_http_error(exc.GitlabListError)
    def list(self, **kwargs):
        """Retrieve a list of objects.

        Args:
            all (bool): If True, return all the items, without pagination
            per_page (int): Number of items to retrieve per request
            page (int): ID of the page to return (starts with page 1)
            as_list (bool): If set to False and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            list: The list of objects, or a generator if `as_list` is False

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the server cannot perform the request
        """

        # Duplicate data to avoid messing with what the user sent us
        data = kwargs.copy()
        if self.gitlab.per_page:
            data.setdefault("per_page", self.gitlab.per_page)

        # global keyset pagination
        if self.gitlab.pagination:
            data.setdefault("pagination", self.gitlab.pagination)

        if self.gitlab.order_by:
            data.setdefault("order_by", self.gitlab.order_by)

        # We get the attributes that need some special transformation
        types = getattr(self, "_types", {})
        if types:
            for attr_name, type_cls in types.items():
                if attr_name in data.keys():
                    type_obj = type_cls(data[attr_name])
                    data[attr_name] = type_obj.get_for_api()

        # Allow to overwrite the path, handy for custom listings
        path = data.pop("path", self.path)

        obj = self.gitlab.http_list(path, **data)
        if isinstance(obj, list):
            return [self._obj_cls(self, item) for item in obj]
        else:
            return base.RESTObjectList(self, self._obj_cls, obj)


class RetrieveMixin(ListMixin, GetMixin):
    pass


class CreateMixin(object):
    def _check_missing_create_attrs(self, data):
        required, optional = self.get_create_attrs()
        missing = []
        for attr in required:
            if attr not in data:
                missing.append(attr)
                continue
        if missing:
            raise AttributeError("Missing attributes: %s" % ", ".join(missing))

    def get_create_attrs(self):
        """Return the required and optional arguments.

        Returns:
            tuple: 2 items: list of required arguments and list of optional
                   arguments for creation (in that order)
        """
        return getattr(self, "_create_attrs", (tuple(), tuple()))

    @exc.on_http_error(exc.GitlabCreateError)
    def create(self, data=None, **kwargs):
        """Create a new object.

        Args:
            data (dict): parameters to send to the server to create the
                         resource
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            RESTObject: a new instance of the managed object class built with
                the data sent by the server

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request
        """
        if data is None:
            data = {}

        self._check_missing_create_attrs(data)
        files = {}

        # We get the attributes that need some special transformation
        types = getattr(self, "_types", {})
        if types:
            # Duplicate data to avoid messing with what the user sent us
            data = data.copy()
            for attr_name, type_cls in types.items():
                if attr_name in data.keys():
                    type_obj = type_cls(data[attr_name])

                    # if the type if FileAttribute we need to pass the data as
                    # file
                    if issubclass(type_cls, g_types.FileAttribute):
                        k = type_obj.get_file_name(attr_name)
                        files[attr_name] = (k, data.pop(attr_name))
                    else:
                        data[attr_name] = type_obj.get_for_api()

        # Handle specific URL for creation
        path = kwargs.pop("path", self.path)
        server_data = self.gitlab.http_post(path, post_data=data, files=files, **kwargs)
        return self._obj_cls(self, server_data)


class UpdateMixin(object):
    def _check_missing_update_attrs(self, data):
        required, optional = self.get_update_attrs()
        # Remove the id field from the required list as it was previously moved to the http path.
        required = tuple(filter(lambda k: k != self._obj_cls._id_attr, required))
        missing = []
        for attr in required:
            if attr not in data:
                missing.append(attr)
                continue
        if missing:
            raise AttributeError("Missing attributes: %s" % ", ".join(missing))

    def get_update_attrs(self):
        """Return the required and optional arguments.

        Returns:
            tuple: 2 items: list of required arguments and list of optional
                   arguments for update (in that order)
        """
        return getattr(self, "_update_attrs", (tuple(), tuple()))

    def _get_update_method(self):
        """Return the HTTP method to use.

        Returns:
            object: http_put (default) or http_post
        """
        if getattr(self, "_update_uses_post", False):
            http_method = self.gitlab.http_post
        else:
            http_method = self.gitlab.http_put
        return http_method

    @exc.on_http_error(exc.GitlabUpdateError)
    def update(self, id=None, new_data=None, **kwargs):
        """Update an object on the server.

        Args:
            id: ID of the object to update (can be None if not required)
            new_data: the update data for the object
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            dict: The new object data (*not* a RESTObject)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        new_data = new_data or {}

        if id is None:
            path = self.path
        else:
            path = "%s/%s" % (self.path, id)

        self._check_missing_update_attrs(new_data)
        files = {}

        # We get the attributes that need some special transformation
        types = getattr(self, "_types", {})
        if types:
            # Duplicate data to avoid messing with what the user sent us
            new_data = new_data.copy()
            for attr_name, type_cls in types.items():
                if attr_name in new_data.keys():
                    type_obj = type_cls(new_data[attr_name])

                    # if the type if FileAttribute we need to pass the data as
                    # file
                    if issubclass(type_cls, g_types.FileAttribute):
                        k = type_obj.get_file_name(attr_name)
                        files[attr_name] = (k, new_data.pop(attr_name))
                    else:
                        new_data[attr_name] = type_obj.get_for_api()

        http_method = self._get_update_method()
        return http_method(path, post_data=new_data, files=files, **kwargs)


class SetMixin(object):
    @exc.on_http_error(exc.GitlabSetError)
    def set(self, key, value, **kwargs):
        """Create or update the object.

        Args:
            key (str): The key of the object to create/update
            value (str): The value to set for the object
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSetError: If an error occured

        Returns:
            obj: The created/updated attribute
        """
        path = "%s/%s" % (self.path, utils.clean_str_id(key))
        data = {"value": value}
        server_data = self.gitlab.http_put(path, post_data=data, **kwargs)
        return self._obj_cls(self, server_data)


class DeleteMixin(object):
    @exc.on_http_error(exc.GitlabDeleteError)
    def delete(self, id, **kwargs):
        """Delete an object on the server.

        Args:
            id: ID of the object to delete
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        if id is None:
            path = self.path
        else:
            if not isinstance(id, int):
                id = utils.clean_str_id(id)
            path = "%s/%s" % (self.path, id)
        self.gitlab.http_delete(path, **kwargs)


class CRUDMixin(GetMixin, ListMixin, CreateMixin, UpdateMixin, DeleteMixin):
    pass


class NoUpdateMixin(GetMixin, ListMixin, CreateMixin, DeleteMixin):
    pass


class SaveMixin(object):
    """Mixin for RESTObject's that can be updated."""

    def _get_updated_data(self):
        updated_data = {}
        required, optional = self.manager.get_update_attrs()
        for attr in required:
            # Get everything required, no matter if it's been updated
            updated_data[attr] = getattr(self, attr)
        # Add the updated attributes
        updated_data.update(self._updated_attrs)

        return updated_data

    def save(self, **kwargs):
        """Save the changes made to the object to the server.

        The object is updated to match what the server returns.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raise:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        updated_data = self._get_updated_data()
        # Nothing to update. Server fails if sent an empty dict.
        if not updated_data:
            return

        # call the manager
        obj_id = self.get_id()
        server_data = self.manager.update(obj_id, updated_data, **kwargs)
        if server_data is not None:
            self._update_attrs(server_data)


class ObjectDeleteMixin(object):
    """Mixin for RESTObject's that can be deleted."""

    def delete(self, **kwargs):
        """Delete the object from the server.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        self.manager.delete(self.get_id())


class UserAgentDetailMixin(object):
    @cli.register_custom_action(("Snippet", "ProjectSnippet", "ProjectIssue"))
    @exc.on_http_error(exc.GitlabGetError)
    def user_agent_detail(self, **kwargs):
        """Get the user agent detail.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request
        """
        path = "%s/%s/user_agent_detail" % (self.manager.path, self.get_id())
        return self.manager.gitlab.http_get(path, **kwargs)


class AccessRequestMixin(object):
    @cli.register_custom_action(
        ("ProjectAccessRequest", "GroupAccessRequest"), tuple(), ("access_level",)
    )
    @exc.on_http_error(exc.GitlabUpdateError)
    def approve(self, access_level=gitlab.DEVELOPER_ACCESS, **kwargs):
        """Approve an access request.

        Args:
            access_level (int): The access level for the user
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server fails to perform the request
        """

        path = "%s/%s/approve" % (self.manager.path, self.id)
        data = {"access_level": access_level}
        server_data = self.manager.gitlab.http_put(path, post_data=data, **kwargs)
        self._update_attrs(server_data)


class DownloadMixin(object):
    @cli.register_custom_action(("GroupExport", "ProjectExport"))
    @exc.on_http_error(exc.GitlabGetError)
    def download(self, streamed=False, action=None, chunk_size=1024, **kwargs):
        """Download the archive of a resource export.

        Args:
            streamed (bool): If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                reatment
            action (callable): Callable responsible of dealing with chunk of
                data
            chunk_size (int): Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server failed to perform the request

        Returns:
            str: The blob content if streamed is False, None otherwise
        """
        path = "%s/download" % (self.manager.path)
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        return utils.response_content(result, streamed, action, chunk_size)


class SubscribableMixin(object):
    @cli.register_custom_action(
        ("ProjectIssue", "ProjectMergeRequest", "ProjectLabel", "GroupLabel")
    )
    @exc.on_http_error(exc.GitlabSubscribeError)
    def subscribe(self, **kwargs):
        """Subscribe to the object notifications.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSubscribeError: If the subscription cannot be done
        """
        path = "%s/%s/subscribe" % (self.manager.path, self.get_id())
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        self._update_attrs(server_data)

    @cli.register_custom_action(
        ("ProjectIssue", "ProjectMergeRequest", "ProjectLabel", "GroupLabel")
    )
    @exc.on_http_error(exc.GitlabUnsubscribeError)
    def unsubscribe(self, **kwargs):
        """Unsubscribe from the object notifications.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUnsubscribeError: If the unsubscription cannot be done
        """
        path = "%s/%s/unsubscribe" % (self.manager.path, self.get_id())
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        self._update_attrs(server_data)


class TodoMixin(object):
    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
    @exc.on_http_error(exc.GitlabTodoError)
    def todo(self, **kwargs):
        """Create a todo associated to the object.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTodoError: If the todo cannot be set
        """
        path = "%s/%s/todo" % (self.manager.path, self.get_id())
        self.manager.gitlab.http_post(path, **kwargs)


class TimeTrackingMixin(object):
    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
    @exc.on_http_error(exc.GitlabTimeTrackingError)
    def time_stats(self, **kwargs):
        """Get time stats for the object.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTimeTrackingError: If the time tracking update cannot be done
        """
        # Use the existing time_stats attribute if it exist, otherwise make an
        # API call
        if "time_stats" in self.attributes:
            return self.attributes["time_stats"]

        path = "%s/%s/time_stats" % (self.manager.path, self.get_id())
        return self.manager.gitlab.http_get(path, **kwargs)

    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"), ("duration",))
    @exc.on_http_error(exc.GitlabTimeTrackingError)
    def time_estimate(self, duration, **kwargs):
        """Set an estimated time of work for the object.

        Args:
            duration (str): Duration in human format (e.g. 3h30)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTimeTrackingError: If the time tracking update cannot be done
        """
        path = "%s/%s/time_estimate" % (self.manager.path, self.get_id())
        data = {"duration": duration}
        return self.manager.gitlab.http_post(path, post_data=data, **kwargs)

    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
    @exc.on_http_error(exc.GitlabTimeTrackingError)
    def reset_time_estimate(self, **kwargs):
        """Resets estimated time for the object to 0 seconds.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTimeTrackingError: If the time tracking update cannot be done
        """
        path = "%s/%s/reset_time_estimate" % (self.manager.path, self.get_id())
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"), ("duration",))
    @exc.on_http_error(exc.GitlabTimeTrackingError)
    def add_spent_time(self, duration, **kwargs):
        """Add time spent working on the object.

        Args:
            duration (str): Duration in human format (e.g. 3h30)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTimeTrackingError: If the time tracking update cannot be done
        """
        path = "%s/%s/add_spent_time" % (self.manager.path, self.get_id())
        data = {"duration": duration}
        return self.manager.gitlab.http_post(path, post_data=data, **kwargs)

    @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
    @exc.on_http_error(exc.GitlabTimeTrackingError)
    def reset_spent_time(self, **kwargs):
        """Resets the time spent working on the object.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabTimeTrackingError: If the time tracking update cannot be done
        """
        path = "%s/%s/reset_spent_time" % (self.manager.path, self.get_id())
        return self.manager.gitlab.http_post(path, **kwargs)


class ParticipantsMixin(object):
    @cli.register_custom_action(("ProjectMergeRequest", "ProjectIssue"))
    @exc.on_http_error(exc.GitlabListError)
    def participants(self, **kwargs):
        """List the participants.

        Args:
            all (bool): If True, return all the items, without pagination
            per_page (int): Number of items to retrieve per request
            page (int): ID of the page to return (starts with page 1)
            as_list (bool): If set to False and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the list could not be retrieved

        Returns:
            RESTObjectList: The list of participants
        """

        path = "%s/%s/participants" % (self.manager.path, self.get_id())
        return self.manager.gitlab.http_get(path, **kwargs)


class BadgeRenderMixin(object):
    @cli.register_custom_action(
        ("GroupBadgeManager", "ProjectBadgeManager"), ("link_url", "image_url")
    )
    @exc.on_http_error(exc.GitlabRenderError)
    def render(self, link_url, image_url, **kwargs):
        """Preview link_url and image_url after interpolation.

        Args:
            link_url (str): URL of the badge link
            image_url (str): URL of the badge image
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabRenderError: If the rendering failed

        Returns:
            dict: The rendering properties
        """
        path = "%s/render" % self.path
        data = {"link_url": link_url, "image_url": image_url}
        return self.gitlab.http_get(path, data, **kwargs)



import pytest
import requests

from gitlab import Gitlab


def test_invalid_auth_args():
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            oauth_token="bearer",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            oauth_token="bearer",
            http_username="foo",
            http_password="bar",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            http_password="bar",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            http_username="foo",
        )


def test_private_token_auth():
    gl = Gitlab("http://localhost", private_token="private_token", api_version="4")
    assert gl.private_token == "private_token"
    assert gl.oauth_token is None
    assert gl.job_token is None
    assert gl._http_auth is None
    assert "Authorization" not in gl.headers
    assert gl.headers["PRIVATE-TOKEN"] == "private_token"
    assert "JOB-TOKEN" not in gl.headers


def test_oauth_token_auth():
    gl = Gitlab("http://localhost", oauth_token="oauth_token", api_version="4")
    assert gl.private_token is None
    assert gl.oauth_token == "oauth_token"
    assert gl.job_token is None
    assert gl._http_auth is None
    assert gl.headers["Authorization"] == "Bearer oauth_token"
    assert "PRIVATE-TOKEN" not in gl.headers
    assert "JOB-TOKEN" not in gl.headers


def test_job_token_auth():
    gl = Gitlab("http://localhost", job_token="CI_JOB_TOKEN", api_version="4")
    assert gl.private_token is None
    assert gl.oauth_token is None
    assert gl.job_token == "CI_JOB_TOKEN"
    assert gl._http_auth is None
    assert "Authorization" not in gl.headers
    assert "PRIVATE-TOKEN" not in gl.headers
    assert gl.headers["JOB-TOKEN"] == "CI_JOB_TOKEN"


def test_http_auth():
    gl = Gitlab(
        "http://localhost",
        private_token="private_token",
        http_username="foo",
        http_password="bar",
        api_version="4",
    )
    assert gl.private_token == "private_token"
    assert gl.oauth_token is None
    assert gl.job_token is None
    assert isinstance(gl._http_auth, requests.auth.HTTPBasicAuth)
    assert gl.headers["PRIVATE-TOKEN"] == "private_token"
    assert "Authorization" not in gl.headers



"""
GitLab API: https://docs.gitlab.com/ee/api/resource_milestone_events.html
"""

import pytest
import responses

from gitlab.v4.objects import (
    ProjectIssueResourceMilestoneEvent,
    ProjectMergeRequestResourceMilestoneEvent,
)


@pytest.fixture()
def resp_merge_request_milestone_events():
    mr_content = {"iid": 1}
    events_content = {"id": 1, "resource_type": "MergeRequest"}
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests",
            json=[mr_content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/resource_milestone_events",
            json=[events_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_project_issue_milestone_events():
    issue_content = {"iid": 1}
    events_content = {"id": 1, "resource_type": "Issue"}
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues",
            json=[issue_content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/issues/1/resource_milestone_events",
            json=[events_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_project_issue_milestone_events(project, resp_project_issue_milestone_events):
    issue = project.issues.list()[0]
    milestone_events = issue.resourcemilestoneevents.list()
    assert isinstance(milestone_events, list)
    milestone_event = milestone_events[0]
    assert isinstance(milestone_event, ProjectIssueResourceMilestoneEvent)
    assert milestone_event.resource_type == "Issue"


def test_merge_request_milestone_events(project, resp_merge_request_milestone_events):
    mr = project.mergerequests.list()[0]
    milestone_events = mr.resourcemilestoneevents.list()
    assert isinstance(milestone_events, list)
    milestone_event = milestone_events[0]
    assert isinstance(milestone_event, ProjectMergeRequestResourceMilestoneEvent)
    assert milestone_event.resource_type == "MergeRequest"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import operator
import sys

import gitlab
import gitlab.base
from gitlab import cli
import gitlab.v4.objects


class GitlabCLI(object):
    def __init__(self, gl, what, action, args):
        self.cls_name = cli.what_to_cls(what)
        self.cls = gitlab.v4.objects.__dict__[self.cls_name]
        self.what = what.replace("-", "_")
        self.action = action.lower()
        self.gl = gl
        self.args = args
        self.mgr_cls = getattr(gitlab.v4.objects, self.cls.__name__ + "Manager")
        # We could do something smart, like splitting the manager name to find
        # parents, build the chain of managers to get to the final object.
        # Instead we do something ugly and efficient: interpolate variables in
        # the class _path attribute, and replace the value with the result.
        self.mgr_cls._path = self.mgr_cls._path % self.args
        self.mgr = self.mgr_cls(gl)

        types = getattr(self.mgr_cls, "_types", {})
        if types:
            for attr_name, type_cls in types.items():
                if attr_name in self.args.keys():
                    obj = type_cls()
                    obj.set_from_cli(self.args[attr_name])
                    self.args[attr_name] = obj.get()

    def __call__(self):
        # Check for a method that matches object + action
        method = "do_%s_%s" % (self.what, self.action)
        if hasattr(self, method):
            return getattr(self, method)()

        # Fallback to standard actions (get, list, create, ...)
        method = "do_%s" % self.action
        if hasattr(self, method):
            return getattr(self, method)()

        # Finally try to find custom methods
        return self.do_custom()

    def do_custom(self):
        in_obj = cli.custom_actions[self.cls_name][self.action][2]

        # Get the object (lazy), then act
        if in_obj:
            data = {}
            if hasattr(self.mgr, "_from_parent_attrs"):
                for k in self.mgr._from_parent_attrs:
                    data[k] = self.args[k]
            if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(self.cls):
                data[self.cls._id_attr] = self.args.pop(self.cls._id_attr)
            o = self.cls(self.mgr, data)
            method_name = self.action.replace("-", "_")
            return getattr(o, method_name)(**self.args)
        else:
            return getattr(self.mgr, self.action)(**self.args)

    def do_project_export_download(self):
        try:
            project = self.gl.projects.get(int(self.args["project_id"]), lazy=True)
            data = project.exports.get().download()
            if hasattr(sys.stdout, "buffer"):
                # python3
                sys.stdout.buffer.write(data)
            else:
                sys.stdout.write(data)

        except Exception as e:
            cli.die("Impossible to download the export", e)

    def do_create(self):
        try:
            return self.mgr.create(self.args)
        except Exception as e:
            cli.die("Impossible to create object", e)

    def do_list(self):
        try:
            return self.mgr.list(**self.args)
        except Exception as e:
            cli.die("Impossible to list objects", e)

    def do_get(self):
        id = None
        if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(self.mgr_cls):
            id = self.args.pop(self.cls._id_attr)

        try:
            return self.mgr.get(id, **self.args)
        except Exception as e:
            cli.die("Impossible to get object", e)

    def do_delete(self):
        id = self.args.pop(self.cls._id_attr)
        try:
            self.mgr.delete(id, **self.args)
        except Exception as e:
            cli.die("Impossible to destroy object", e)

    def do_update(self):
        id = None
        if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(self.mgr_cls):
            id = self.args.pop(self.cls._id_attr)
        try:
            return self.mgr.update(id, self.args)
        except Exception as e:
            cli.die("Impossible to update object", e)


def _populate_sub_parser_by_class(cls, sub_parser):
    mgr_cls_name = cls.__name__ + "Manager"
    mgr_cls = getattr(gitlab.v4.objects, mgr_cls_name)

    for action_name in ["list", "get", "create", "update", "delete"]:
        if not hasattr(mgr_cls, action_name):
            continue

        sub_parser_action = sub_parser.add_parser(action_name)
        sub_parser_action.add_argument("--sudo", required=False)
        if hasattr(mgr_cls, "_from_parent_attrs"):
            [
                sub_parser_action.add_argument(
                    "--%s" % x.replace("_", "-"), required=True
                )
                for x in mgr_cls._from_parent_attrs
            ]

        if action_name == "list":
            if hasattr(mgr_cls, "_list_filters"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=False
                    )
                    for x in mgr_cls._list_filters
                ]

            sub_parser_action.add_argument("--page", required=False)
            sub_parser_action.add_argument("--per-page", required=False)
            sub_parser_action.add_argument("--all", required=False, action="store_true")

        if action_name == "delete":
            if cls._id_attr is not None:
                id_attr = cls._id_attr.replace("_", "-")
                sub_parser_action.add_argument("--%s" % id_attr, required=True)

        if action_name == "get":
            if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(cls):
                if cls._id_attr is not None:
                    id_attr = cls._id_attr.replace("_", "-")
                    sub_parser_action.add_argument("--%s" % id_attr, required=True)

            if hasattr(mgr_cls, "_optional_get_attrs"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=False
                    )
                    for x in mgr_cls._optional_get_attrs
                ]

        if action_name == "create":
            if hasattr(mgr_cls, "_create_attrs"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=True
                    )
                    for x in mgr_cls._create_attrs[0]
                ]

                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=False
                    )
                    for x in mgr_cls._create_attrs[1]
                ]

        if action_name == "update":
            if cls._id_attr is not None:
                id_attr = cls._id_attr.replace("_", "-")
                sub_parser_action.add_argument("--%s" % id_attr, required=True)

            if hasattr(mgr_cls, "_update_attrs"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=True
                    )
                    for x in mgr_cls._update_attrs[0]
                    if x != cls._id_attr
                ]

                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=False
                    )
                    for x in mgr_cls._update_attrs[1]
                    if x != cls._id_attr
                ]

    if cls.__name__ in cli.custom_actions:
        name = cls.__name__
        for action_name in cli.custom_actions[name]:
            sub_parser_action = sub_parser.add_parser(action_name)
            # Get the attributes for URL/path construction
            if hasattr(mgr_cls, "_from_parent_attrs"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=True
                    )
                    for x in mgr_cls._from_parent_attrs
                ]
                sub_parser_action.add_argument("--sudo", required=False)

            # We need to get the object somehow
            if gitlab.mixins.GetWithoutIdMixin not in inspect.getmro(cls):
                if cls._id_attr is not None:
                    id_attr = cls._id_attr.replace("_", "-")
                    sub_parser_action.add_argument("--%s" % id_attr, required=True)

            required, optional, dummy = cli.custom_actions[name][action_name]
            [
                sub_parser_action.add_argument(
                    "--%s" % x.replace("_", "-"), required=True
                )
                for x in required
                if x != cls._id_attr
            ]
            [
                sub_parser_action.add_argument(
                    "--%s" % x.replace("_", "-"), required=False
                )
                for x in optional
                if x != cls._id_attr
            ]

    if mgr_cls.__name__ in cli.custom_actions:
        name = mgr_cls.__name__
        for action_name in cli.custom_actions[name]:
            sub_parser_action = sub_parser.add_parser(action_name)
            if hasattr(mgr_cls, "_from_parent_attrs"):
                [
                    sub_parser_action.add_argument(
                        "--%s" % x.replace("_", "-"), required=True
                    )
                    for x in mgr_cls._from_parent_attrs
                ]
                sub_parser_action.add_argument("--sudo", required=False)

            required, optional, dummy = cli.custom_actions[name][action_name]
            [
                sub_parser_action.add_argument(
                    "--%s" % x.replace("_", "-"), required=True
                )
                for x in required
                if x != cls._id_attr
            ]
            [
                sub_parser_action.add_argument(
                    "--%s" % x.replace("_", "-"), required=False
                )
                for x in optional
                if x != cls._id_attr
            ]


def extend_parser(parser):
    subparsers = parser.add_subparsers(
        title="object", dest="what", help="Object to manipulate."
    )
    subparsers.required = True

    # populate argparse for all Gitlab Object
    classes = []
    for cls in gitlab.v4.objects.__dict__.values():
        try:
            if gitlab.base.RESTManager in inspect.getmro(cls):
                if cls._obj_cls is not None:
                    classes.append(cls._obj_cls)
        except AttributeError:
            pass
    classes.sort(key=operator.attrgetter("__name__"))

    for cls in classes:
        arg_name = cli.cls_to_what(cls)
        object_group = subparsers.add_parser(arg_name)

        object_subparsers = object_group.add_subparsers(
            title="action", dest="whaction", help="Action to execute."
        )
        _populate_sub_parser_by_class(cls, object_subparsers)
        object_subparsers.required = True

    return parser


def get_dict(obj, fields):
    if isinstance(obj, str):
        return obj

    if fields:
        return {k: v for k, v in obj.attributes.items() if k in fields}
    return obj.attributes


class JSONPrinter(object):
    def display(self, d, **kwargs):
        import json  # noqa

        print(json.dumps(d))

    def display_list(self, data, fields, **kwargs):
        import json  # noqa

        print(json.dumps([get_dict(obj, fields) for obj in data]))


class YAMLPrinter(object):
    def display(self, d, **kwargs):
        try:
            import yaml  # noqa

            print(yaml.safe_dump(d, default_flow_style=False))
        except ImportError:
            exit(
                "PyYaml is not installed.\n"
                "Install it with `pip install PyYaml` "
                "to use the yaml output feature"
            )

    def display_list(self, data, fields, **kwargs):
        try:
            import yaml  # noqa

            print(
                yaml.safe_dump(
                    [get_dict(obj, fields) for obj in data], default_flow_style=False
                )
            )
        except ImportError:
            exit(
                "PyYaml is not installed.\n"
                "Install it with `pip install PyYaml` "
                "to use the yaml output feature"
            )


class LegacyPrinter(object):
    def display(self, d, **kwargs):
        verbose = kwargs.get("verbose", False)
        padding = kwargs.get("padding", 0)
        obj = kwargs.get("obj")

        def display_dict(d, padding):
            for k in sorted(d.keys()):
                v = d[k]
                if isinstance(v, dict):
                    print("%s%s:" % (" " * padding, k.replace("_", "-")))
                    new_padding = padding + 2
                    self.display(v, verbose=True, padding=new_padding, obj=v)
                    continue
                print("%s%s: %s" % (" " * padding, k.replace("_", "-"), v))

        if verbose:
            if isinstance(obj, dict):
                display_dict(obj, padding)
                return

            # not a dict, we assume it's a RESTObject
            if obj._id_attr:
                id = getattr(obj, obj._id_attr, None)
                print("%s: %s" % (obj._id_attr, id))
            attrs = obj.attributes
            if obj._id_attr:
                attrs.pop(obj._id_attr)
            display_dict(attrs, padding)

        else:
            if obj._id_attr:
                id = getattr(obj, obj._id_attr)
                print("%s: %s" % (obj._id_attr.replace("_", "-"), id))
            if hasattr(obj, "_short_print_attr"):
                value = getattr(obj, obj._short_print_attr) or "None"
                value = value.replace("\r", "").replace("\n", " ")
                # If the attribute is a note (ProjectCommitComment) then we do
                # some modifications to fit everything on one line
                line = "%s: %s" % (obj._short_print_attr, value)
                # ellipsize long lines (comments)
                if len(line) > 79:
                    line = line[:76] + "..."
                print(line)

    def display_list(self, data, fields, **kwargs):
        verbose = kwargs.get("verbose", False)
        for obj in data:
            if isinstance(obj, gitlab.base.RESTObject):
                self.display(get_dict(obj, fields), verbose=verbose, obj=obj)
            else:
                print(obj)
            print("")


PRINTERS = {"json": JSONPrinter, "legacy": LegacyPrinter, "yaml": YAMLPrinter}


def run(gl, what, action, args, verbose, output, fields):
    g_cli = GitlabCLI(gl, what, action, args)
    data = g_cli()

    printer = PRINTERS[output]()

    if isinstance(data, dict):
        printer.display(data, verbose=True, obj=data)
    elif isinstance(data, list):
        printer.display_list(data, fields, verbose=verbose)
    elif isinstance(data, gitlab.base.RESTObject):
        printer.display(get_dict(data, fields), verbose=verbose, obj=data)
    elif isinstance(data, str):
        print(data)
    elif hasattr(data, "decode"):
        print(data.decode())



import os
import time


def test_create_project(gitlab_cli):
    name = "test-project1"

    cmd = ["project", "create", "--name", name]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert name in ret.stdout


def test_update_project(gitlab_cli, project):
    description = "My New Description"

    cmd = ["project", "update", "--id", project.id, "--description", description]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert description in ret.stdout


def test_create_group(gitlab_cli):
    name = "test-group1"
    path = "group1"

    cmd = ["group", "create", "--name", name, "--path", path]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert name in ret.stdout
    assert path in ret.stdout


def test_update_group(gitlab_cli, gl, group):
    description = "My New Description"

    cmd = ["group", "update", "--id", group.id, "--description", description]
    ret = gitlab_cli(cmd)

    assert ret.success

    group = gl.groups.list(description=description)[0]
    assert group.description == description


def test_create_user(gitlab_cli, gl):
    email = "fake@email.com"
    username = "user1"
    name = "User One"
    password = "fakepassword"

    cmd = [
        "user",
        "create",
        "--email",
        email,
        "--username",
        username,
        "--name",
        name,
        "--password",
        password,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success

    user = gl.users.list(username=username)[0]

    assert user.email == email
    assert user.username == username
    assert user.name == name


def test_get_user_by_id(gitlab_cli, user):
    cmd = ["user", "get", "--id", user.id]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert str(user.id) in ret.stdout


def test_list_users_verbose_output(gitlab_cli):
    cmd = ["-v", "user", "list"]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert "avatar-url" in ret.stdout


def test_cli_args_not_in_output(gitlab_cli):
    cmd = ["-v", "user", "list"]
    ret = gitlab_cli(cmd)

    assert "config-file" not in ret.stdout


def test_add_member_to_project(gitlab_cli, project, user):
    access_level = "40"

    cmd = [
        "project-member",
        "create",
        "--project-id",
        project.id,
        "--user-id",
        user.id,
        "--access-level",
        access_level,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_user_memberships(gitlab_cli, user):
    cmd = ["user-membership", "list", "--user-id", user.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_project_create_file(gitlab_cli, project):
    file_path = "README"
    branch = "master"
    content = "CONTENT"
    commit_message = "Initial commit"

    cmd = [
        "project-file",
        "create",
        "--project-id",
        project.id,
        "--file-path",
        file_path,
        "--branch",
        branch,
        "--content",
        content,
        "--commit-message",
        commit_message,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_project_issue(gitlab_cli, project):
    title = "my issue"
    description = "my issue description"

    cmd = [
        "project-issue",
        "create",
        "--project-id",
        project.id,
        "--title",
        title,
        "--description",
        description,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert title in ret.stdout


def test_create_issue_note(gitlab_cli, issue):
    body = "body"

    cmd = [
        "project-issue-note",
        "create",
        "--project-id",
        issue.project_id,
        "--issue-iid",
        issue.iid,
        "--body",
        body,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_branch(gitlab_cli, project):
    branch = "branch1"

    cmd = [
        "project-branch",
        "create",
        "--project-id",
        project.id,
        "--branch",
        branch,
        "--ref",
        "master",
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_merge_request(gitlab_cli, project):
    branch = "branch1"

    cmd = [
        "project-merge-request",
        "create",
        "--project-id",
        project.id,
        "--source-branch",
        branch,
        "--target-branch",
        "master",
        "--title",
        "Update README",
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_accept_request_merge(gitlab_cli, project):
    # MR needs at least 1 commit before we can merge
    mr = project.mergerequests.list()[0]
    file_data = {
        "branch": mr.source_branch,
        "file_path": "README2",
        "content": "Content",
        "commit_message": "Pre-merge commit",
    }
    project.files.create(file_data)
    time.sleep(2)

    cmd = [
        "project-merge-request",
        "merge",
        "--project-id",
        project.id,
        "--iid",
        mr.iid,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_revert_commit(gitlab_cli, project):
    commit = project.commits.list()[0]

    cmd = [
        "project-commit",
        "revert",
        "--project-id",
        project.id,
        "--id",
        commit.id,
        "--branch",
        "master",
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_get_commit_signature_not_found(gitlab_cli, project):
    commit = project.commits.list()[0]

    cmd = ["project-commit", "signature", "--project-id", project.id, "--id", commit.id]
    ret = gitlab_cli(cmd)

    assert not ret.success
    assert "404 Signature Not Found" in ret.stderr


def test_create_project_label(gitlab_cli, project):
    name = "prjlabel1"
    description = "prjlabel1 description"
    color = "#112233"

    cmd = [
        "-v",
        "project-label",
        "create",
        "--project-id",
        project.id,
        "--name",
        name,
        "--description",
        description,
        "--color",
        color,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_project_labels(gitlab_cli, project):
    cmd = ["-v", "project-label", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_update_project_label(gitlab_cli, label):
    new_label = "prjlabel2"
    new_description = "prjlabel2 description"
    new_color = "#332211"

    cmd = [
        "-v",
        "project-label",
        "update",
        "--project-id",
        label.project_id,
        "--name",
        label.name,
        "--new-name",
        new_label,
        "--description",
        new_description,
        "--color",
        new_color,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_delete_project_label(gitlab_cli, label):
    # TODO: due to update above, we'd need a function-scope label fixture
    label_name = "prjlabel2"

    cmd = [
        "-v",
        "project-label",
        "delete",
        "--project-id",
        label.project_id,
        "--name",
        label_name,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_group_label(gitlab_cli, group):
    name = "grouplabel1"
    description = "grouplabel1 description"
    color = "#112233"

    cmd = [
        "-v",
        "group-label",
        "create",
        "--group-id",
        group.id,
        "--name",
        name,
        "--description",
        description,
        "--color",
        color,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_labels(gitlab_cli, group):
    cmd = ["-v", "group-label", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_update_group_label(gitlab_cli, group_label):
    new_label = "grouplabel2"
    new_description = "grouplabel2 description"
    new_color = "#332211"

    cmd = [
        "-v",
        "group-label",
        "update",
        "--group-id",
        group_label.group_id,
        "--name",
        group_label.name,
        "--new-name",
        new_label,
        "--description",
        new_description,
        "--color",
        new_color,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_delete_group_label(gitlab_cli, group_label):
    # TODO: due to update above, we'd need a function-scope label fixture
    new_label = "grouplabel2"

    cmd = [
        "-v",
        "group-label",
        "delete",
        "--group-id",
        group_label.group_id,
        "--name",
        new_label,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_project_variable(gitlab_cli, project):
    key = "junk"
    value = "car"

    cmd = [
        "-v",
        "project-variable",
        "create",
        "--project-id",
        project.id,
        "--key",
        key,
        "--value",
        value,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_get_project_variable(gitlab_cli, variable):
    cmd = [
        "-v",
        "project-variable",
        "get",
        "--project-id",
        variable.project_id,
        "--key",
        variable.key,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_update_project_variable(gitlab_cli, variable):
    new_value = "bus"

    cmd = [
        "-v",
        "project-variable",
        "update",
        "--project-id",
        variable.project_id,
        "--key",
        variable.key,
        "--value",
        new_value,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_project_variables(gitlab_cli, project):
    cmd = ["-v", "project-variable", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_delete_project_variable(gitlab_cli, variable):
    cmd = [
        "-v",
        "project-variable",
        "delete",
        "--project-id",
        variable.project_id,
        "--key",
        variable.key,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_delete_branch(gitlab_cli, project):
    # TODO: branch fixture
    branch = "branch1"

    cmd = ["project-branch", "delete", "--project-id", project.id, "--name", branch]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_project_upload_file(gitlab_cli, project):
    cmd = [
        "project",
        "upload",
        "--id",
        project.id,
        "--filename",
        __file__,
        "--filepath",
        os.path.realpath(__file__),
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_get_application_settings(gitlab_cli):
    cmd = ["application-settings", "get"]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_update_application_settings(gitlab_cli):
    cmd = ["application-settings", "update", "--signup-enabled", "false"]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_project_with_values_from_file(gitlab_cli, tmpdir):
    name = "gitlab-project-from-file"
    description = "Multiline\n\nData\n"
    from_file = tmpdir.join(name)
    from_file.write(description)
    from_file_path = f"@{str(from_file)}"

    cmd = [
        "-v",
        "project",
        "create",
        "--name",
        name,
        "--description",
        from_file_path,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert description in ret.stdout


def test_create_project_deploy_token(gitlab_cli, project):
    name = "project-token"
    username = "root"
    expires_at = "2021-09-09"
    scopes = "read_registry"

    cmd = [
        "-v",
        "project-deploy-token",
        "create",
        "--project-id",
        project.id,
        "--name",
        name,
        "--username",
        username,
        "--expires-at",
        expires_at,
        "--scopes",
        scopes,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert name in ret.stdout
    assert username in ret.stdout
    assert expires_at in ret.stdout
    assert scopes in ret.stdout


def test_list_all_deploy_tokens(gitlab_cli, deploy_token):
    cmd = ["-v", "deploy-token", "list"]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert deploy_token.name in ret.stdout
    assert str(deploy_token.id) in ret.stdout
    assert deploy_token.username in ret.stdout
    assert deploy_token.expires_at in ret.stdout
    assert deploy_token.scopes[0] in ret.stdout


def test_list_project_deploy_tokens(gitlab_cli, deploy_token):
    cmd = [
        "-v",
        "project-deploy-token",
        "list",
        "--project-id",
        deploy_token.project_id,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert deploy_token.name in ret.stdout
    assert str(deploy_token.id) in ret.stdout
    assert deploy_token.username in ret.stdout
    assert deploy_token.expires_at in ret.stdout
    assert deploy_token.scopes[0] in ret.stdout


def test_delete_project_deploy_token(gitlab_cli, deploy_token):
    cmd = [
        "-v",
        "project-deploy-token",
        "delete",
        "--project-id",
        deploy_token.project_id,
        "--id",
        deploy_token.id,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    # TODO assert not in list


def test_create_group_deploy_token(gitlab_cli, group):
    name = "group-token"
    username = "root"
    expires_at = "2021-09-09"
    scopes = "read_registry"

    cmd = [
        "-v",
        "group-deploy-token",
        "create",
        "--group-id",
        group.id,
        "--name",
        name,
        "--username",
        username,
        "--expires-at",
        expires_at,
        "--scopes",
        scopes,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert name in ret.stdout
    assert username in ret.stdout
    assert expires_at in ret.stdout
    assert scopes in ret.stdout


def test_list_group_deploy_tokens(gitlab_cli, group_deploy_token):
    cmd = [
        "-v",
        "group-deploy-token",
        "list",
        "--group-id",
        group_deploy_token.group_id,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    assert group_deploy_token.name in ret.stdout
    assert str(group_deploy_token.id) in ret.stdout
    assert group_deploy_token.username in ret.stdout
    assert group_deploy_token.expires_at in ret.stdout
    assert group_deploy_token.scopes[0] in ret.stdout


def test_delete_group_deploy_token(gitlab_cli, group_deploy_token):
    cmd = [
        "-v",
        "group-deploy-token",
        "delete",
        "--group-id",
        group_deploy_token.group_id,
        "--id",
        group_deploy_token.id,
    ]
    ret = gitlab_cli(cmd)

    assert ret.success
    # TODO assert not in list


def test_delete_project(gitlab_cli, project):
    cmd = ["project", "delete", "--id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_delete_group(gitlab_cli, group):
    cmd = ["group", "delete", "--id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success



import pytest


@pytest.fixture
def gitlab_cli(script_runner, CONFIG):
    """Wrapper fixture to help make test cases less verbose."""

    def _gitlab_cli(subcommands):
        """
        Return a script_runner.run method that takes a default gitlab
        command, and subcommands passed as arguments inside test cases.
        """
        command = ["gitlab", "--config-file", CONFIG]

        for subcommand in subcommands:
            # ensure we get strings (e.g from IDs)
            command.append(str(subcommand))

        return script_runner.run(*command)

    return _gitlab_cli



#!/usr/bin/python
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages


def get_version():
    with open("gitlab/__init__.py") as f:
        for line in f:
            if line.startswith("__version__"):
                return eval(line.split("=")[-1])


with open("README.rst", "r") as readme_file:
    readme = readme_file.read()

setup(
    name="python-gitlab",
    version=get_version(),
    description="Interact with GitLab API",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="Gauvain Pocentek",
    author_email="gauvain@pocentek.net",
    license="LGPLv3",
    url="https://github.com/python-gitlab/python-gitlab",
    packages=find_packages(),
    install_requires=["requests>=2.22.0"],
    python_requires=">=3.6.0",
    entry_points={"console_scripts": ["gitlab = gitlab.cli:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    extras_require={
        "autocompletion": ["argcomplete>=1.10.0,<2"],
        "yaml": ["PyYaml>=5.2"],
    },
)



#!/usr/bin/env python

import gitlab


P1 = "root/project1"
P2 = "root/project2"
MR_P1 = 1
I_P1 = 1
I_P2 = 1
EPIC_ISSUES = [4, 5]
G1 = "group1"
LDAP_CN = "app1"
LDAP_PROVIDER = "ldapmain"


def start_log(message):
    print("Testing %s... " % message, end="")


def end_log():
    print("OK")


gl = gitlab.Gitlab.from_config("ee")
project1 = gl.projects.get(P1)
project2 = gl.projects.get(P2)
issue_p1 = project1.issues.get(I_P1)
issue_p2 = project2.issues.get(I_P2)
group1 = gl.groups.get(G1)
mr = project1.mergerequests.get(1)

start_log("MR approvals")
approval = project1.approvals.get()
v = approval.reset_approvals_on_push
approval.reset_approvals_on_push = not v
approval.save()
approval = project1.approvals.get()
assert v != approval.reset_approvals_on_push
project1.approvals.set_approvers(1, [1], [])
approval = project1.approvals.get()
assert approval.approvers[0]["user"]["id"] == 1

approval = mr.approvals.get()
approval.approvals_required = 2
approval.save()
approval = mr.approvals.get()
assert approval.approvals_required == 2
approval.approvals_required = 3
approval.save()
approval = mr.approvals.get()
assert approval.approvals_required == 3
mr.approvals.set_approvers(1, [1], [])
approval = mr.approvals.get()
assert approval.approvers[0]["user"]["id"] == 1

ars = project1.approvalrules.list(all=True)
assert len(ars) == 0
project1.approvalrules.create(
    {"name": "approval-rule", "approvals_required": 1, "group_ids": [group1.id]}
)
ars = project1.approvalrules.list(all=True)
assert len(ars) == 1
assert ars[0].approvals_required == 2
ars[0].save()
ars = project1.approvalrules.list(all=True)
assert len(ars) == 1
assert ars[0].approvals_required == 2
ars[0].delete()
ars = project1.approvalrules.list(all=True)
assert len(ars) == 0
end_log()

start_log("geo nodes")
# very basic tests because we only have 1 node...
nodes = gl.geonodes.list()
status = gl.geonodes.status()
end_log()

start_log("issue links")
# bit of cleanup just in case
for link in issue_p1.links.list():
    issue_p1.links.delete(link.issue_link_id)

src, dst = issue_p1.links.create({"target_project_id": P2, "target_issue_iid": I_P2})
links = issue_p1.links.list()
link_id = links[0].issue_link_id
issue_p1.links.delete(link_id)
end_log()

start_log("LDAP links")
# bit of cleanup just in case
if hasattr(group1, "ldap_group_links"):
    for link in group1.ldap_group_links:
        group1.delete_ldap_group_link(link["cn"], link["provider"])
assert gl.ldapgroups.list()
group1.add_ldap_group_link(LDAP_CN, 30, LDAP_PROVIDER)
group1.ldap_sync()
group1.delete_ldap_group_link(LDAP_CN)
end_log()

start_log("boards")
# bit of cleanup just in case
for board in project1.boards.list():
    if board.name == "testboard":
        board.delete()
board = project1.boards.create({"name": "testboard"})
board = project1.boards.get(board.id)
project1.boards.delete(board.id)

for board in group1.boards.list():
    if board.name == "testboard":
        board.delete()
board = group1.boards.create({"name": "testboard"})
board = group1.boards.get(board.id)
group1.boards.delete(board.id)
end_log()

start_log("push rules")
pr = project1.pushrules.get()
if pr:
    pr.delete()
pr = project1.pushrules.create({"deny_delete_tag": True})
pr.deny_delete_tag = False
pr.save()
pr = project1.pushrules.get()
assert pr is not None
assert pr.deny_delete_tag == False
pr.delete()
end_log()

start_log("license")
l = gl.get_license()
assert "user_limit" in l
try:
    gl.set_license("dummykey")
except Exception as e:
    assert "The license key is invalid." in e.error_message
end_log()

start_log("epics")
epic = group1.epics.create({"title": "Test epic"})
epic.title = "Fixed title"
epic.labels = ["label1", "label2"]
epic.save()
epic = group1.epics.get(epic.iid)
assert epic.title == "Fixed title"
assert len(group1.epics.list())

# issues
assert not epic.issues.list()
for i in EPIC_ISSUES:
    epic.issues.create({"issue_id": i})
assert len(EPIC_ISSUES) == len(epic.issues.list())
for ei in epic.issues.list():
    ei.delete()

epic.delete()
end_log()



"""
GitLab API: https://docs.gitlab.com/ce/api/applications.html
"""

import pytest
import responses


title = "GitLab Test Instance"
description = "gitlab-test.example.com"
new_title = "new-title"
new_description = "new-description"


@pytest.fixture
def resp_application_create():
    content = {
        "name": "test_app",
        "redirect_uri": "http://localhost:8080",
        "scopes": ["api", "email"],
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/applications",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_create_application(gl, resp_application_create):
    application = gl.applications.create(
        {
            "name": "test_app",
            "redirect_uri": "http://localhost:8080",
            "scopes": ["api", "email"],
            "confidential": False,
        }
    )
    assert application.name == "test_app"
    assert application.redirect_uri == "http://localhost:8080"
    assert application.scopes == ["api", "email"]



"""
GitLab API: https://docs.gitlab.com/ce/api/system_hooks.html
"""
import pytest
import responses

from gitlab.v4.objects import Hook


@pytest.fixture
def resp_get_hook():
    content = {"url": "testurl", "id": 1}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/hooks/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_hooks(gl, resp_get_hook):
    data = gl.hooks.get(1)
    assert isinstance(data, Hook)
    assert data.url == "testurl"
    assert data.id == 1



"""
GitLab API: https://docs.gitlab.com/ce/api/deployments.html
"""
import pytest
import responses


@pytest.fixture
def resp_deployment():
    content = {"id": 42, "status": "success", "ref": "master"}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/deployments",
            json=content,
            content_type="application/json",
            status=200,
        )

        updated_content = dict(content)
        updated_content["status"] = "failed"

        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/deployments/42",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_deployment(project, resp_deployment):
    deployment = project.deployments.create(
        {
            "environment": "Test",
            "sha": "1agf4gs",
            "ref": "master",
            "tag": False,
            "status": "created",
        }
    )
    assert deployment.id == 42
    assert deployment.status == "success"
    assert deployment.ref == "master"

    deployment.status = "failed"
    deployment.save()
    assert deployment.status == "failed"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools


class GitlabError(Exception):
    def __init__(self, error_message="", response_code=None, response_body=None):

        Exception.__init__(self, error_message)
        # Http status code
        self.response_code = response_code
        # Full http response
        self.response_body = response_body
        # Parsed error message from gitlab
        try:
            # if we receive str/bytes we try to convert to unicode/str to have
            # consistent message types (see #616)
            self.error_message = error_message.decode()
        except Exception:
            self.error_message = error_message

    def __str__(self):
        if self.response_code is not None:
            return "{0}: {1}".format(self.response_code, self.error_message)
        else:
            return "{0}".format(self.error_message)


class GitlabAuthenticationError(GitlabError):
    pass


class RedirectError(GitlabError):
    pass


class GitlabParsingError(GitlabError):
    pass


class GitlabConnectionError(GitlabError):
    pass


class GitlabOperationError(GitlabError):
    pass


class GitlabHttpError(GitlabError):
    pass


class GitlabListError(GitlabOperationError):
    pass


class GitlabGetError(GitlabOperationError):
    pass


class GitlabCreateError(GitlabOperationError):
    pass


class GitlabUpdateError(GitlabOperationError):
    pass


class GitlabDeleteError(GitlabOperationError):
    pass


class GitlabSetError(GitlabOperationError):
    pass


class GitlabProtectError(GitlabOperationError):
    pass


class GitlabTransferProjectError(GitlabOperationError):
    pass


class GitlabProjectDeployKeyError(GitlabOperationError):
    pass


class GitlabCancelError(GitlabOperationError):
    pass


class GitlabPipelineCancelError(GitlabCancelError):
    pass


class GitlabRetryError(GitlabOperationError):
    pass


class GitlabBuildCancelError(GitlabCancelError):
    pass


class GitlabBuildRetryError(GitlabRetryError):
    pass


class GitlabBuildPlayError(GitlabRetryError):
    pass


class GitlabBuildEraseError(GitlabRetryError):
    pass


class GitlabJobCancelError(GitlabCancelError):
    pass


class GitlabJobRetryError(GitlabRetryError):
    pass


class GitlabJobPlayError(GitlabRetryError):
    pass


class GitlabJobEraseError(GitlabRetryError):
    pass


class GitlabPipelinePlayError(GitlabRetryError):
    pass


class GitlabPipelineRetryError(GitlabRetryError):
    pass


class GitlabBlockError(GitlabOperationError):
    pass


class GitlabUnblockError(GitlabOperationError):
    pass


class GitlabDeactivateError(GitlabOperationError):
    pass


class GitlabActivateError(GitlabOperationError):
    pass


class GitlabSubscribeError(GitlabOperationError):
    pass


class GitlabUnsubscribeError(GitlabOperationError):
    pass


class GitlabMRForbiddenError(GitlabOperationError):
    pass


class GitlabMRApprovalError(GitlabOperationError):
    pass


class GitlabMRRebaseError(GitlabOperationError):
    pass


class GitlabMRClosedError(GitlabOperationError):
    pass


class GitlabMROnBuildSuccessError(GitlabOperationError):
    pass


class GitlabTodoError(GitlabOperationError):
    pass


class GitlabTimeTrackingError(GitlabOperationError):
    pass


class GitlabUploadError(GitlabOperationError):
    pass


class GitlabAttachFileError(GitlabOperationError):
    pass


class GitlabImportError(GitlabOperationError):
    pass


class GitlabCherryPickError(GitlabOperationError):
    pass


class GitlabHousekeepingError(GitlabOperationError):
    pass


class GitlabOwnershipError(GitlabOperationError):
    pass


class GitlabSearchError(GitlabOperationError):
    pass


class GitlabStopError(GitlabOperationError):
    pass


class GitlabMarkdownError(GitlabOperationError):
    pass


class GitlabVerifyError(GitlabOperationError):
    pass


class GitlabRenderError(GitlabOperationError):
    pass


class GitlabRepairError(GitlabOperationError):
    pass


class GitlabRevertError(GitlabOperationError):
    pass


class GitlabLicenseError(GitlabOperationError):
    pass


def on_http_error(error):
    """Manage GitlabHttpError exceptions.

    This decorator function can be used to catch GitlabHttpError exceptions
    raise specialized exceptions instead.

    Args:
        error(Exception): The exception type to raise -- must inherit from
            GitlabError
    """

    def wrap(f):
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except GitlabHttpError as e:
                raise error(e.error_message, e.response_code, e.response_body) from e

        return wrapped_f

    return wrap



import re

import pytest
import responses

import gitlab


runner_detail = {
    "active": True,
    "architecture": "amd64",
    "description": "test-1-20150125",
    "id": 6,
    "ip_address": "127.0.0.1",
    "is_shared": False,
    "contacted_at": "2016-01-25T16:39:48.066Z",
    "name": "test-runner",
    "online": True,
    "status": "online",
    "platform": "linux",
    "projects": [
        {
            "id": 1,
            "name": "GitLab Community Edition",
            "name_with_namespace": "GitLab.org / GitLab Community Edition",
            "path": "gitlab-foss",
            "path_with_namespace": "gitlab-org/gitlab-foss",
        }
    ],
    "revision": "5nj35",
    "tag_list": ["ruby", "mysql"],
    "version": "v13.0.0",
    "access_level": "ref_protected",
    "maximum_timeout": 3600,
}

runner_shortinfo = {
    "active": True,
    "description": "test-1-20150125",
    "id": 6,
    "is_shared": False,
    "ip_address": "127.0.0.1",
    "name": "test-name",
    "online": True,
    "status": "online",
}

runner_jobs = [
    {
        "id": 6,
        "ip_address": "127.0.0.1",
        "status": "running",
        "stage": "test",
        "name": "test",
        "ref": "master",
        "tag": False,
        "coverage": "99%",
        "created_at": "2017-11-16T08:50:29.000Z",
        "started_at": "2017-11-16T08:51:29.000Z",
        "finished_at": "2017-11-16T08:53:29.000Z",
        "duration": 120,
        "user": {
            "id": 1,
            "name": "John Doe2",
            "username": "user2",
            "state": "active",
            "avatar_url": "http://www.gravatar.com/avatar/c922747a93b40d1ea88262bf1aebee62?s=80&d=identicon",
            "web_url": "http://localhost/user2",
            "created_at": "2017-11-16T18:38:46.000Z",
            "bio": None,
            "location": None,
            "public_email": "",
            "skype": "",
            "linkedin": "",
            "twitter": "",
            "website_url": "",
            "organization": None,
        },
    }
]


@pytest.fixture
def resp_get_runners_jobs():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/runners/6/jobs",
            json=runner_jobs,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_runners_list():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r".*?(/runners(/all)?|/(groups|projects)/1/runners)"),
            json=[runner_shortinfo],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_runner_detail():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?/runners/6")
        rsps.add(
            method=responses.GET,
            url=pattern,
            json=runner_detail,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.PUT,
            url=pattern,
            json=runner_detail,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_runner_register():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?/runners")
        rsps.add(
            method=responses.POST,
            url=pattern,
            json={"id": "6", "token": "6337ff461c94fd3fa32ba3b1ff4125"},
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_runner_enable():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?(projects|groups)/1/runners")
        rsps.add(
            method=responses.POST,
            url=pattern,
            json=runner_shortinfo,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_runner_delete():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?/runners/6")
        rsps.add(
            method=responses.GET,
            url=pattern,
            json=runner_detail,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.DELETE,
            url=pattern,
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_runner_disable():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?/(groups|projects)/1/runners/6")
        rsps.add(
            method=responses.DELETE,
            url=pattern,
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_runner_verify():
    with responses.RequestsMock() as rsps:
        pattern = re.compile(r".*?/runners/verify")
        rsps.add(
            method=responses.POST,
            url=pattern,
            status=200,
        )
        yield rsps


def test_owned_runners_list(gl: gitlab.Gitlab, resp_get_runners_list):
    runners = gl.runners.list()
    assert runners[0].active == True
    assert runners[0].id == 6
    assert runners[0].name == "test-name"
    assert len(runners) == 1


def test_project_runners_list(gl: gitlab.Gitlab, resp_get_runners_list):
    runners = gl.projects.get(1, lazy=True).runners.list()
    assert runners[0].active == True
    assert runners[0].id == 6
    assert runners[0].name == "test-name"
    assert len(runners) == 1


def test_group_runners_list(gl: gitlab.Gitlab, resp_get_runners_list):
    runners = gl.groups.get(1, lazy=True).runners.list()
    assert runners[0].active == True
    assert runners[0].id == 6
    assert runners[0].name == "test-name"
    assert len(runners) == 1


def test_all_runners_list(gl: gitlab.Gitlab, resp_get_runners_list):
    runners = gl.runners.all()
    assert runners[0].active == True
    assert runners[0].id == 6
    assert runners[0].name == "test-name"
    assert len(runners) == 1


def test_create_runner(gl: gitlab.Gitlab, resp_runner_register):
    runner = gl.runners.create({"token": "token"})
    assert runner.id == "6"
    assert runner.token == "6337ff461c94fd3fa32ba3b1ff4125"


def test_get_update_runner(gl: gitlab.Gitlab, resp_runner_detail):
    runner = gl.runners.get(6)
    assert runner.active == True
    runner.tag_list.append("new")
    runner.save()


def test_remove_runner(gl: gitlab.Gitlab, resp_runner_delete):
    runner = gl.runners.get(6)
    runner.delete()
    gl.runners.delete(6)


def test_disable_project_runner(gl: gitlab.Gitlab, resp_runner_disable):
    gl.projects.get(1, lazy=True).runners.delete(6)


def test_disable_group_runner(gl: gitlab.Gitlab, resp_runner_disable):
    gl.groups.get(1, lazy=True).runners.delete(6)


def test_enable_project_runner(gl: gitlab.Gitlab, resp_runner_enable):
    runner = gl.projects.get(1, lazy=True).runners.create({"runner_id": 6})
    assert runner.active == True
    assert runner.id == 6
    assert runner.name == "test-name"


def test_enable_group_runner(gl: gitlab.Gitlab, resp_runner_enable):
    runner = gl.groups.get(1, lazy=True).runners.create({"runner_id": 6})
    assert runner.active == True
    assert runner.id == 6
    assert runner.name == "test-name"


def test_verify_runner(gl: gitlab.Gitlab, resp_runner_verify):
    gl.runners.verify("token")


def test_runner_jobs(gl: gitlab.Gitlab, resp_get_runners_jobs):
    jobs = gl.runners.get(6, lazy=True).jobs.list()
    assert jobs[0].duration == 120
    assert jobs[0].name == "test"
    assert jobs[0].user.get("name") == "John Doe2"
    assert len(jobs) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import argparse
import functools
import importlib
import re
import sys

import gitlab.config

camel_re = re.compile("(.)([A-Z])")

# custom_actions = {
#    cls: {
#        action: (mandatory_args, optional_args, in_obj),
#    },
# }
custom_actions = {}


def register_custom_action(cls_names, mandatory=tuple(), optional=tuple()):
    def wrap(f):
        @functools.wraps(f)
        def wrapped_f(*args, **kwargs):
            return f(*args, **kwargs)

        # in_obj defines whether the method belongs to the obj or the manager
        in_obj = True
        classes = cls_names
        if type(cls_names) != tuple:
            classes = (cls_names,)

        for cls_name in classes:
            final_name = cls_name
            if cls_name.endswith("Manager"):
                final_name = cls_name.replace("Manager", "")
                in_obj = False
            if final_name not in custom_actions:
                custom_actions[final_name] = {}

            action = f.__name__.replace("_", "-")
            custom_actions[final_name][action] = (mandatory, optional, in_obj)

        return wrapped_f

    return wrap


def die(msg, e=None):
    if e:
        msg = "%s (%s)" % (msg, e)
    sys.stderr.write(msg + "\n")
    sys.exit(1)


def what_to_cls(what):
    return "".join([s.capitalize() for s in what.split("-")])


def cls_to_what(cls):
    return camel_re.sub(r"\1-\2", cls.__name__).lower()


def _get_base_parser(add_help=True):
    parser = argparse.ArgumentParser(
        add_help=add_help, description="GitLab API Command Line Interface"
    )
    parser.add_argument("--version", help="Display the version.", action="store_true")
    parser.add_argument(
        "-v",
        "--verbose",
        "--fancy",
        help="Verbose mode (legacy format only)",
        action="store_true",
    )
    parser.add_argument(
        "-d", "--debug", help="Debug mode (display HTTP requests)", action="store_true"
    )
    parser.add_argument(
        "-c",
        "--config-file",
        action="append",
        help="Configuration file to use. Can be used multiple times.",
    )
    parser.add_argument(
        "-g",
        "--gitlab",
        help=(
            "Which configuration section should "
            "be used. If not defined, the default selection "
            "will be used."
        ),
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Output format (v4 only): json|legacy|yaml",
        required=False,
        choices=["json", "legacy", "yaml"],
        default="legacy",
    )
    parser.add_argument(
        "-f",
        "--fields",
        help=(
            "Fields to display in the output (comma "
            "separated). Not used with legacy output"
        ),
        required=False,
    )

    return parser


def _get_parser(cli_module):
    parser = _get_base_parser()
    return cli_module.extend_parser(parser)


def _parse_value(v):
    if isinstance(v, str) and v.startswith("@"):
        # If the user-provided value starts with @, we try to read the file
        # path provided after @ as the real value. Exit on any error.
        try:
            with open(v[1:]) as fl:
                return fl.read()
        except Exception as e:
            sys.stderr.write("%s\n" % e)
            sys.exit(1)

    return v


def main():
    if "--version" in sys.argv:
        print(gitlab.__version__)
        sys.exit(0)

    parser = _get_base_parser(add_help=False)

    # This first parsing step is used to find the gitlab config to use, and
    # load the propermodule (v3 or v4) accordingly. At that point we don't have
    # any subparser setup
    (options, args) = parser.parse_known_args(sys.argv)
    try:
        config = gitlab.config.GitlabConfigParser(options.gitlab, options.config_file)
    except gitlab.config.ConfigError as e:
        if "--help" in sys.argv or "-h" in sys.argv:
            parser.print_help()
            sys.exit(0)
        sys.exit(e)
    cli_module = importlib.import_module("gitlab.v%s.cli" % config.api_version)

    # Now we build the entire set of subcommands and do the complete parsing
    parser = _get_parser(cli_module)
    try:
        import argcomplete

        argcomplete.autocomplete(parser)
    except Exception:
        pass
    args = parser.parse_args(sys.argv[1:])

    config_files = args.config_file
    gitlab_id = args.gitlab
    verbose = args.verbose
    output = args.output
    fields = []
    if args.fields:
        fields = [x.strip() for x in args.fields.split(",")]
    debug = args.debug
    action = args.whaction
    what = args.what

    args = args.__dict__
    # Remove CLI behavior-related args
    for item in (
        "gitlab",
        "config_file",
        "verbose",
        "debug",
        "what",
        "whaction",
        "version",
        "output",
    ):
        args.pop(item)
    args = {k: _parse_value(v) for k, v in args.items() if v is not None}

    try:
        gl = gitlab.Gitlab.from_config(gitlab_id, config_files)
        if gl.private_token or gl.oauth_token or gl.job_token:
            gl.auth()
    except Exception as e:
        die(str(e))

    if debug:
        gl.enable_debug()

    cli_module.run(gl, what, action, args, verbose, output, fields)

    sys.exit(0)



"""
GitLab API: https://docs.gitlab.com/ce/api/deploy_tokens.html
"""
import pytest
import responses

from gitlab.v4.objects import ProjectDeployToken


create_content = {
    "id": 1,
    "name": "test_deploy_token",
    "username": "custom-user",
    "expires_at": "2022-01-01T00:00:00.000Z",
    "token": "jMRvtPNxrn3crTAGukpZ",
    "scopes": ["read_repository"],
}


@pytest.fixture
def resp_deploy_token_create():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/deploy_tokens",
            json=create_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_deploy_tokens(gl, resp_deploy_token_create):
    deploy_token = gl.projects.get(1, lazy=True).deploytokens.create(
        {
            "name": "test_deploy_token",
            "expires_at": "2022-01-01T00:00:00.000Z",
            "username": "custom-user",
            "scopes": ["read_repository"],
        }
    )
    assert isinstance(deploy_token, ProjectDeployToken)
    assert deploy_token.id == 1
    assert deploy_token.expires_at == "2022-01-01T00:00:00.000Z"
    assert deploy_token.username == "custom-user"
    assert deploy_token.scopes == ["read_repository"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 Mika Mäenpää <mika.j.maenpaa@tut.fi>,
#                    Tampere University of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or`
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

from httmock import HTTMock, response, urlmatch, with_httmock  # noqa

from gitlab import Gitlab, GitlabList
from gitlab.v4.objects import CurrentUser


username = "username"
user_id = 1


@urlmatch(scheme="http", netloc="localhost", path="/api/v4/user", method="get")
def resp_get_user(url, request):
    headers = {"content-type": "application/json"}
    content = '{{"id": {0:d}, "username": "{1:s}"}}'.format(user_id, username).encode(
        "utf-8"
    )
    return response(200, content, headers, None, 5, request)


@urlmatch(scheme="http", netloc="localhost", path="/api/v4/tests", method="get")
def resp_page_1(url, request):
    headers = {
        "content-type": "application/json",
        "X-Page": 1,
        "X-Next-Page": 2,
        "X-Per-Page": 1,
        "X-Total-Pages": 2,
        "X-Total": 2,
        "Link": ("<http://localhost/api/v4/tests?per_page=1&page=2>;" ' rel="next"'),
    }
    content = '[{"a": "b"}]'
    return response(200, content, headers, None, 5, request)


@urlmatch(
    scheme="http",
    netloc="localhost",
    path="/api/v4/tests",
    method="get",
    query=r".*page=2",
)
def resp_page_2(url, request):
    headers = {
        "content-type": "application/json",
        "X-Page": 2,
        "X-Next-Page": 2,
        "X-Per-Page": 1,
        "X-Total-Pages": 2,
        "X-Total": 2,
    }
    content = '[{"c": "d"}]'
    return response(200, content, headers, None, 5, request)


def test_gitlab_build_list(gl):
    with HTTMock(resp_page_1):
        obj = gl.http_list("/tests", as_list=False)
    assert len(obj) == 2
    assert obj._next_url == "http://localhost/api/v4/tests?per_page=1&page=2"
    assert obj.current_page == 1
    assert obj.prev_page is None
    assert obj.next_page == 2
    assert obj.per_page == 1
    assert obj.total_pages == 2
    assert obj.total == 2

    with HTTMock(resp_page_2):
        l = list(obj)
    assert len(l) == 2
    assert l[0]["a"] == "b"
    assert l[1]["c"] == "d"


@with_httmock(resp_page_1, resp_page_2)
def test_gitlab_all_omitted_when_as_list(gl):
    result = gl.http_list("/tests", as_list=False, all=True)
    assert isinstance(result, GitlabList)


def test_gitlab_strip_base_url(gl_trailing):
    assert gl_trailing.url == "http://localhost"


def test_gitlab_strip_api_url(gl_trailing):
    assert gl_trailing.api_url == "http://localhost/api/v4"


def test_gitlab_build_url(gl_trailing):
    r = gl_trailing._build_url("/projects")
    assert r == "http://localhost/api/v4/projects"


def test_gitlab_pickability(gl):
    original_gl_objects = gl._objects
    pickled = pickle.dumps(gl)
    unpickled = pickle.loads(pickled)
    assert isinstance(unpickled, Gitlab)
    assert hasattr(unpickled, "_objects")
    assert unpickled._objects == original_gl_objects


@with_httmock(resp_get_user)
def test_gitlab_token_auth(gl, callback=None):
    gl.auth()
    assert gl.user.username == username
    assert gl.user.id == user_id
    assert isinstance(gl.user, CurrentUser)


def test_gitlab_from_config(default_config):
    config_path = default_config
    Gitlab.from_config("one", [config_path])


def test_gitlab_subclass_from_config(default_config):
    class MyGitlab(Gitlab):
        pass

    config_path = default_config
    gl = MyGitlab.from_config("one", [config_path])
    assert isinstance(gl, MyGitlab)
